// Generate hash_vector function

#include <stdlib.h>
#include <assert_ops/libc.h>
#include <bestiola/stdlib_wrap.h>

#include "conf.h"

typedef enum {
	UNKNOWN,
	USE,
	DISCARD
} status;

status component[OUT_BITS][IN_BITS];
static unsigned count_used[OUT_BITS];
_Bool pair[IN_BITS][IN_BITS];

static int out_compar(const void *p1, const void *p2);
static void output_code(void);
static _Bool pair_conflict(status *component, unsigned cur_in);
static void use_bit(status *component, unsigned cur_in);

int main(void) {
	unsigned tryout[OUT_BITS];
	unsigned i = 0;
	for (; i < OUT_BITS; i++)
		tryout[i] = i;

	unsigned cur_in = 0;
	unsigned last_in = IN_BITS - 1;
	for (;; cur_in = (cur_in + 1) % IN_BITS) {
		_Bool new_use = 0;
		unsigned i = 0;
		for (; i < OUT_BITS; i++) {
			unsigned cur_out = tryout[i];
			if (component[cur_out][cur_in] != UNKNOWN)
				continue;
			if (pair_conflict(component[cur_out], cur_in))
				component[cur_out][cur_in] = DISCARD;
			else {
				use_bit(component[cur_out], cur_in);
				count_used[cur_out]++;
				ARRAY_SORT(tryout, out_compar);
				new_use = 1;
				break;
			}
		}

		if (new_use)
			last_in = cur_in;
		else if (cur_in == last_in)
			break;
	}

	output_code();

	return EXIT_SUCCESS;
}

static int out_compar(const void *p1, const void *p2) {
	const unsigned *i1 = p1;
	const unsigned *i2 = p2;
	unsigned count1 = count_used[*i1];
	unsigned count2 = count_used[*i2];
	if (count1 != count2)
		return count1 - count2;
	return *i1 - *i2;
}

static void output_code(void) {
	unsigned in_bytes = (IN_BITS + 7) / 8;
	unsigned out_bytes = (OUT_BITS + 7) / 8;

	assert_printf(
"#ifndef XORIGUER_HASH_VECTOR_H\n\
#define XORIGUER_HASH_VECTOR_H\n\
\n\
enum {\n\
	IN_BITS = %u,\n\
	IN_BYTES = %u,\n\
	OUT_BITS = %u,\n\
	OUT_BYTES = %u,\n\
	BITS = IN_BITS,\n\
	BYTES = IN_BYTES\n\
};\n\
\n\
static void hash_vector(unsigned char o[OUT_BYTES], unsigned char i[IN_BYTES])\
 {\n", IN_BITS, in_bytes, OUT_BITS, out_bytes);

	unsigned i = 0;
	for (; i < IN_BITS; i++)
		assert_printf("	unsigned char i%d = i[%d] >> %d & 1;\n", i,
			i / 8, i % 8);
	assert_putchar('\n');
	unsigned o;
	for (o = 0; o < OUT_BITS; o++) {
		unsigned byte = o / 8;
		unsigned msb = byte != out_bytes - 1? 7: (OUT_BITS + 7) % 8;
		unsigned bit = msb - o % 8;
		unsigned new_o = byte * 8 + bit;
		if (bit == msb)
			assert_printf("	o[%d] =", byte);
		else
			assert_printf("	o[%d] <<= 1;\n	o[%d] |=", byte, byte);
		_Bool first = 1;
		unsigned i;
		for (i = 0; i < IN_BITS; i++) {
			if (component[new_o][i] == USE) {
				if (first) {
					first = 0;
					assert_printf(" i%d", i);
				}
				else
					assert_printf(" ^ i%d", i);
			}
		}
		if (!new_o)
			assert_printf(" ^ 1");
		assert_puts(";");
	}
	assert_puts("}\n\
\n\
#endif");
}

static _Bool pair_conflict(status *component, unsigned cur_in) {
	unsigned i = 0;
	for (; i < IN_BITS; i++) {
		if (component[i] == USE && pair[cur_in][i])
			return 1;
	}
	return 0;
}

static void use_bit(status *component, unsigned cur_in) {
	unsigned i = 0;
	for (; i < IN_BITS; i++) {
		if (component[i] == USE) {
			pair[cur_in][i] = 1;
			pair[i][cur_in] = 1;
		}
	}
	component[cur_in] = USE;
}
