// Check whether all vectors are unique

#include <stdlib.h>
#include <string.h>
#include <assert_ops/libc.h>
#include <bestiola/assert.h>
#include <bestiola/err_wrap.h>

#include "hash_vector.h"

int main(void) {
	if (IN_BITS != OUT_BITS)
		bad_assertion();

	unsigned char in[BYTES];
	unsigned char out[BYTES];

	size_t END = (size_t) 1 << BITS;

	size_t END8 = BITS >= 3? END / 8: 1;
	size_t MEM_LIMIT = 0x08000000;
	size_t bsize = END8 < MEM_LIMIT? END8: MEM_LIMIT;
	unsigned char *bitmap = assert_malloc(bsize);

	unsigned times = END8 / bsize;
	unsigned time = 0;
	for (; time < times; time++) {
		memset(bitmap, 0, bsize);
		size_t size = 0;
		for (; size < END; size++) {
			unsigned i = 0;
			for (; i < BYTES; i++)
				in[i] = size >> 8 * i & 255;
			hash_vector(out, in);
			size_t o_size = out[0];
			for (i = 1; i < BYTES; i++)
				o_size |= (size_t) out[i] << 8 * i;
			size_t offset = o_size / 8;
			if (offset / bsize != time)
				continue;
			offset %= bsize;
			unsigned bit = o_size % 8;
			if (bitmap[offset] & 1 << bit)
				errx_fail("Failed! %zx", size);
			bitmap[offset] |= 1 << bit;
		}

		unsigned b = 0;
		for (; b < bsize; b++) {
			unsigned char all_set = BITS >= 3? 0xff:
				(1 << 2 * BITS) - 1;
			if (bitmap[b] != all_set)
				errx_fail("Failed! Time %u", time);
		}
		assert_printf("Time %u\n", time);
	}

	free(bitmap);

	assert_puts("Passed");
	return EXIT_SUCCESS;
}
