// Calculate XORiguer sum

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert_ops/libc.h>
#include <bestiola/assert.h>
#include <bestiola/err_wrap.h>

#include "hash_vector.h"
#include "require.h"

static size_t get_size_len(size_t size);
static void hash_block(unsigned char out[BYTES], unsigned char in[BYTES]);

int main(void) {
	REQUIRE_MIX_PREVIOUS;
	REQUIRE_READ_BYTES;
	REQUIRE_SIZE_FITS_IN_BLOCK;

	unsigned char in[BYTES];
	unsigned char out[BYTES];

	memset(out, 0, sizeof out);

	size_t size = 0;
	size_t nread = fread(in, 1, BYTES, stdin);
	for (; nread == BYTES; nread = fread(in, 1, BYTES, stdin)) {
		hash_block(out, in);
		if (size > SIZE_MAX - BYTES)
			errx_fail("Too much input");
		size += BYTES;
	}
	if (ferror(stdin))
		errx_fail("Input error");

	size_t size_len = get_size_len(size);
	size_t si = 0;
	for (; si < size_len && nread < BYTES; si++, nread++) {
		unsigned char size_char = size >> 8 * si;
		in[nread] = size_char;
	}
	memset(in + nread, 0, sizeof in - nread);
	hash_block(out, in);
	if (si < size_len) {
		nread = 0;
		for (; si < size_len; si++, nread++) {
			unsigned char size_char = size >> 8 * si;
			in[nread] = size_char;
		}
		memset(in + nread, 0, sizeof in - nread);
		hash_block(out, in);
	}

	unsigned i = 0;
	for (; i < BYTES; i++)
		assert_printf("%02x", out[i]);
	assert_putchar('\n');

	return EXIT_SUCCESS;
}

static int clzs(size_t x) {
	switch (sizeof x) {
	case sizeof(long):
		return __builtin_clzl(x);
	default:
		bad_assertion();
	}
}

static size_t get_size_len(size_t size) {
	return size? sizeof size - clzs(size) / 8: 0;
}

static void hash_block(unsigned char out[BYTES], unsigned char in[BYTES]) {
	unsigned char hvin[BYTES];
	unsigned char hvout[BYTES];

	hash_vector(hvin, in);
	hash_vector(hvout, out);
	unsigned i = 0;
	for (; i < BYTES; i++)
		out[i] = hvin[i] ^ hvout[(i + BYTES / 2) % BYTES];
}
